<?php
/**
 *
 */
class plgHikashoppaymentPaywithamazon extends hikashopPaymentPlugin {
	var $multiple = true;
	var $name = 'paywithamazon';

	// Currencies code :
	var $accepted_currencies = array('USD', 'EUR',  'JPY', ' CNY');

	// Amazon url :
	// https://amzn.github.io/login-and-pay-with-amazon-sdk-samples/simple.html
	var $urls = array(
		'us'=> array(
			'live_widget' => 'https://static-na.payments-amazon.com/OffAmazonPayments/us/js/Widgets.js',
			'sandbox_widget' => 'https://static-na.payments-amazon.com/OffAmazonPayments/us/sandbox/js/Widgets.js',
			'live_server' => 'https://mws.amazonservices.com/OffAmazonPayments/2013-01-01',
			'sandbox_server' => 'https://mws.amazonservices.com/OffAmazonPayments_Sandbox/2013-01-01',
		),
		'uk' => array (
			'live_widget' => 'https://static-eu.payments-amazon.com/OffAmazonPayments/uk/lpa/js/Widgets.js',
			'sandbox_widget' => 'https://static-eu.payments-amazon.com/OffAmazonPayments/uk/sandbox/lpa/js/Widgets.js',
			'live_server' => 'https://mws-eu.amazonservices.com/OffAmazonPayments/2013-01-01/',
			'sandbox_server' => 'https://mws-eu.amazonservices.com/OffAmazonPayments_Sandbox/2013-01-01'
		),
		'ue' => array (
			'live_widget' => 'https://static-eu.payments-amazon.com/OffAmazonPayments/eur/lpa/js/Widgets.js',
			'sandbox_widget' => 'https://static-eu.payments-amazon.com/OffAmazonPayments/eur/sandbox/lpa/js/Widgets.js',
			'live_server' => 'https://mws-eu.amazonservices.com/OffAmazonPayments/2013-01-01/',
			'sandbox_server' => 'https://mws-eu.amazonservices.com/OffAmazonPayments_Sandbox/2013-01-01'
		),
		'de' => array(
			'live_widget' => 'https://static-eu.payments-amazon.com/OffAmazonPayments/uk/lpa/js/Widgets.js',
			'sandbox_widget' => 'https://static-eu.payments-amazon.com/OffAmazonPayments/uk/sandbox/lpa/js/Widgets.js',
			'live_server' => 'https://mws-eu.amazonservices.com/OffAmazonPayments/2013-01-01/',
			'sandbox_server' => 'https://mws-eu.amazonservices.com/OffAmazonPayments_Sandbox/2013-01-01',
		),
		'jp' => array(
			'live_widget' => 'https://origin-na.ssl-images-amazon.com/images/G/09/EP/offAmazonPayments/prod/lpa/js/Widgets.js',
			'sandbox_widget' => 'https://origin-na.ssl-images-amazon.com/images/G/09/EP/offAmazonPayments/sandbox/prod/lpa/js/Widgets.js',
			'live_server' => 'https://mws.amazonservices.jp/OffAmazonPayments/2013-01-01',
			'sandbox_server' => 'https://mws.amazonservices.jp/OffAmazonPayments_Sandbox/2013-01-01',
		),
	);

	//backend spec config
	var $pluginConfig = array(
		'sellerId' => array('Seller Id', 'input'),
		// required to sign the XMLcart
		'accessKey' => array('Access Key', 'input'),
		'secretAccessKey' => array('Secret Access Key', 'input'),
		'lwaClientId' => array('Client Id', 'input'), //REQUIRED ?...
		// Redirect mode :
		'redirectMode' => array('redirect mode', 'list', array(
			'false' => 'redirection',
			'true' => 'popup',
		)),
		// for Amazon account settings :
		'html1' => array('For amazon notification settings','html',''),
		// Location with required endpoints :
		'account_zone' => array('Account Zone', 'list', array(
			'us' => 'United States',
			'ue' => 'European Union',
			'uk' => 'United Kingdom',
			'de' => 'Deutschland',
			'jp' => 'Japan'
		)),

		'sellerNote' => array('message for customer', 'input'),
		// Btn Settings :
		'color' => array('Amazon button color', 'list', array(
			'Gold' => 'Gold',
			'LightGray' => 'Light Gray',
			'DarkGray' => 'Dark Gray'
		)),
		'size' => array('Amazon button size', 'list', array(
			'small' => 'Small',
			'medium' => 'Medium',
			'large' => 'Large',
			'x-large' => 'X-Large'
		)),

		'debug' => array('DEBUG', 'boolean', '0'),
		'environnement' => array('ENVIRONNEMENT', 'list', array(
			'live' => 'HIKA_PRODUCTION',
			'sandbox' => 'HIKA_SANDBOX'
		)),
		'invalid_status' => array('INVALID_STATUS', 'orderstatus'),
		'verified_status' => array('VERIFIED_STATUS', 'orderstatus'),
		// 'return_url' => array('RETURN_URL', 'html'), For customer redirection
	);

	function __construct(&$subject, $config) {
		$this->pluginConfig['html1'][2] = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=checkout&task=notify&skip_system_notification=1&notif_payment=paywithamazon';

		return parent::__construct($subject, $config);
	}

	/**
	 * Collect data & process with the checkout :
	 */
	public function onAfterOrderConfirm(&$order,&$methods,$method_id){
		parent::onAfterOrderConfirm($order, $methods, $method_id);

		// Backend data check :
		if (empty ($this->payment_params->sellerId) ) {
			$this->app->enqueueMessage('You have to configure a Seller ID for the Pay with Amazon plugin payment first : check your plugin\'s parameters, on your website backend','error');
			return false;
		}
		if (empty($this->payment_params->accessKey)) {
			$this->app->enqueueMessage('You have to configure the Access Key for the Pay with Amazon plugin payment first : check your plugin\'s parameters, on your website backend','error');
			return false;
		}
		if (empty($this->payment_params->lwaClientId)) {
			$this->app->enqueueMessage('You have to configure the LwaClient Id for the Pay with Amazon plugin payment first : check your plugin\'s parameters, on your website backend','error');
			return false;
		}

		//
		//
		if(empty($this->payment_params->account_zone) || !in_array($this->payment_params->account_zone, array('us','eu','jp')))
			$this->payment_params->account_zone = 'us';

		// Required amazon url via merchant Location / depends from location AND mode (prod/Sndbx):
		$amazonUrl = @$this->urls[$this->payment_params->account_zone][$this->payment_params->environnement.'_widget'];

		$current_locale = $this->locale;
		if(!in_array($current_locale, array('en-GB','en-US','fr-FR')))
			$current_locale = 'en-US';

		// Set required data for prepare the Amazon button (log) :
		//
		$vars = array(
			'sellerId' => trim($this->payment_params->sellerId),
			'accessKey' => trim($this->payment_params->accessKey),
			'lwaClientId' => trim($this->payment_params->lwaClientId),
			'currencyCode' => $this->currency->currency_code,
			'sellerNote' => trim($this->payment_params->sellerNote),
			// Btn display setting :
			'color' => $this->payment_params->color,
			'size' => $this->payment_params->size,
			// Customer Langage settings
			'language' => $current_locale,
			// Redirect mode :
			'redirectMode' => $this->payment_params->redirectMode,
			// Notify page :
			'returnUrl' => HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=checkout&task=notify&skip_system_notification=1&notif_payment=paywithamazon',
			// Sandbox / Production :
			'amazonUrl' => $amazonUrl
		);

		// For debug in log file :
		if($this->payment_params->debug) {
			$this->writeToLog("Data sent to Pay with Amazon: \n\n");
			$this->writeToLog(print_r($vars,true));
		}
		// Create Amazon Button :
		$AmazonBtn = $this->AmazonBtn($vars);

		// Define Amazon Url for the request :
		$this->payment_params->url = $amazonUrl;
		$this->vars = $AmazonBtn;

		return $this->showPage('end');
	}

	/**
	 * Treatment start withPay With Amazon response /POST:
	 */
	public function onPaymentNotification(&$statuses){
		//Load the parameters of the plugin
		$app = JFactory::getApplication();
		$order_id = $app->getUserState('com_hikashop.order_id');
		$dbOrder = $this->getOrder($order_id);
		$this->loadPaymentParams($dbOrder);
		if(empty($this->payment_params))
			return false;

		// WALLET WIDGETS
		if(!isset($_POST['payment_selected'])) {
			//Debug check:
			if($this->payment_params->debug) {
				$this->writeToLog("Data recieved from Amazon after login :  \n\n");
				$this->writeToLog(print_r($_GET,true));
			}

			// Amazon Url setting :
			// Required amazon url via merchant Location / depends from location AND mode (prod/Sndbx):
			$amazonUrl = $this->urls[$this->payment_params->account_zone][$this->payment_params->environnement.'_widget'];
			// Second return to onPaymentNofification :
			$returnNotif = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=checkout&task=notify&notif_payment=paywithamazon';

			// SET vars for Wallet widget view :
			$vars = array (
				'returnNotif' => $returnNotif,
				'access_token' => $_GET["access_token"],
				'sellerId' => $this->payment_params->sellerId,
				'lwaClientId' => $this->payment_params->lwaClientId,
				'amazonUrl' => $amazonUrl
			);

			// Create Wallet html & javascript :
			$this->vars = $vars;

			ob_start();
			$this->showPage('wallet');
			$data = ob_get_clean();
			return $data;
		}
		// FOR Check/flag
		$check = true;

		// Treatment after Wallet widget :
		if($this->payment_params->debug) {
			$this->writeToLog("Data recieved from Amazon after WALLET : \n\n ");
			$this->writeToLog(print_r($_GET,true));
		}

		// Get Currency :
		$currencyClass = hikashop_get('class.currency');
		$currencies = $currencyClass->getCurrencies($dbOrder->order_currency_id,$currencies);
		$currency = $currencies[$dbOrder->order_currency_id];

		// Check Required data to proceed API Calls :
		if (!isset($_POST['reference_id']) || !isset($_POST['access_token']) || !isset($currency)) {
			if($this->payment_params->debug) {
				$this->writeToLog('Empty required data for continu : \n\n');
				$this->writeToLog('currency value : /n/n');
				$this->writeToLog($currency);
				$this->writeToLog('Acces token value : /n/n');
				$this->writeToLog($_POST['access_token']);
				$this->writeToLog('Reference id value : /n/n');
				$this->writeToLog($_POST['reference_id']);
			}
			$check = false;
		}
		else {
			// Define Url parameters for API calls :
			$amazonUrl = $this->urls[$this->payment_params->account_zone][$this->payment_params->environnement.'_server'];
			$parsed_url = $this->setUrlArray($amazonUrl);

			// Set API call required :
			$context = array(
				'reference_id' => $_POST['reference_id'],
				'access_token' => $_POST['access_token'],
				'currency' => $currency,
			);
			// SET ORDER REFERENCE DETAILS
			// REQUEST PARAMS
			$params = $this->getArrayParams('SetOrderReferenceDetails', $dbOrder, $context);
			// Set Params & add signature :
			$stringData = $this->getSignedData($params);

			// Send request
			$response = $this->sendRequest($parsed_url, $stringData);

			// Continu Check :
			if ($response == false)
				$check = $response;

			unset($params);
			unset($response);
			unset($stringData);
		}
		// CONFIRM ORDER REFERENCE
		if ($check == true) {
			//  REQUEST PARAMS
			$params = $this->getArrayParams('ConfirmOrderReference', $dbOrder, $context);
			// Set Params & add signature :
			$stringData = $this->getSignedData($params);

			// Send request (Add token for curl request)
			$response = $this->sendRequest($parsed_url, $stringData);

			// Continu Check :
			if ($response == false)
				$check = false;

			unset($params);
			unset($response);
			unset($stringData);
		}
		// AUTHORIZE
		if ($check == true) {
			// REQUEST PARAMS :
			$params = $this->getArrayParams('Authorize', $dbOrder, $context);
			// Set Params & add signature :
			$stringData = $this->getSignedData($params);

			// Send request
			$response = $this->sendRequest($parsed_url, $stringData);

			// Continu Check :
			if ($response == false)
				$check = false;

			unset($params);
			unset($stringData);

			// Response object treatment :
			$RespArr = simplexml_load_string($response);
		}
		// CLOSED
		if ($check == true) {
			// REQUEST PARAMS :
			$params = $this->getArrayParams('CloseOrderReference', $dbOrder, $context);
			// Set Params & add signature :
			$stringData = $this->getSignedData($params);

			// Send request
			$response = $this->sendRequest($parsed_url, $stringData);
		}

		// Get order id from Amazon notification :
		$amazonOrder_id = (string) $RespArr->AuthorizeResult->AuthorizationDetails->AuthorizationReferenceId;

		// Order already completed ?
		if (($dbOrder->order_status == 'confirmed') && ($check == true)) {
			if($this->payment_params->debug){
				echo 'Pay with Amazon HikaShop plg : order already paid'."\n\n\n";
				echo 'Order Id :'. $amazonOrder_id . '\n\n\n';
			}
			// End process here & redirect to failed page :
			$EndUrl = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=order&task=cancel_order&order_id=' . $dbOrder->order_id . $this->url_itemid;
			$app->enqueueMessage('Transaction Failed');
			$app->redirect($EndUrl);

			return false;
		}
		if ($check == true) {
			// Get data from amazon notification results :
			$amazonAmount = (string) $RespArr->AuthorizeResult->AuthorizationDetails->AuthorizationAmount->Amount;
			$amazonCurrency = (string) $RespArr->AuthorizeResult->AuthorizationDetails->AuthorizationAmount->CurrencyCode;

			// Amazon Data comparison WITH our Db order check :
			if ($amazonAmount != round($dbOrder->order_full_price, 2)) {
				$check = false;
				$DataName = 'Amount';
				$badData = $amazonAmount;
				$ourData = round($dbOrder->order_full_price, 2);
			}
			if (($amazonCurrency != $currency->currency_code) && ($check == true)) {
				$check = false;
				$DataName = 'Currency';
				$badData = $amazonCurrency;
				$ourData = $currency;
			}
			if (($amazonOrder_id != $dbOrder->order_id) && ($check == true)) {
				$check = false;
				$DataName = 'Order id';
				$badData = $amazonOrder_id;
				$ourData = $dbOrder->order_id;
			}
		}
		// Order treatment :
		$order_text = "\r\n" . JText::sprintf('NOTIFICATION_OF_ORDER_ON_WEBSITE', $dbOrder->order_number, HIKASHOP_LIVE);
		$order_text .= "\r\n" . str_replace('<br/>', "\r\n", JText::sprintf('ACCESS_ORDER_WITH_LINK', $url));

		if ($check == false) {
			// Transaction Failed :
			$EndUrl = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=order&task=cancel_order&order_id=' . $dbOrder->order_id . $this->url_itemid;

			if(($this->payment_params->debug) && (isset($ourData))) {
				$this->writeToLog('invalid response from Pay with Amazon'."\n\n\n");
				$this->writeToLog(''.$DataName." don't match \n\n\n");
				$this->writeToLog('Expected : '.$ourData." recieved :" .$badData. "\n\n\n");
			}
			// Details for Order history :
			$details = 'Amazon reference Id : '. $context['reference_id'];

			$email = new stdClass();
			$email->subject = JText::sprintf('NOTIFICATION_REFUSED_FOR_THE_ORDER', $this->name).'invalid response';
			$email->body = JText::sprintf("Hello,\r\n A Pay with Amazon notification was refused because the response from the Amazon server was invalid")."\r\n\r\n".$order_text;
			$history = new stdClass();
			$history->data = $details;

			// Order update :
			$this->modifyOrder($dbOrder->order_id, $this->payment_params->invalid_status, $history, $email);

			$app->enqueueMessage('Transaction Failed');
			$app->redirect($EndUrl);

			return false;
		}
		else {
			// Details for Order history :
			$details = 'Amazon reference Id : '. $context['reference_id'];

			// SUCCESSFUL transaction :
			// History
			$history = new stdClass();
			$history->notified = 1;
			$history->amount = $amazonAmount;
			$history->data = $details;

			// Email
			$email = new stdClass();
			$email->subject = JText::sprintf('PAYMENT_NOTIFICATION_FOR_ORDER', 'Amazon', 'Success', $dbOrder->order_number);
			$body = str_replace('<br/>', "\r\n", JText::sprintf('PAYMENT_NOTIFICATION_STATUS', 'Amazon', 'Success')) . ' ' . JText::sprintf('ORDER_STATUS_CHANGED', $this->payment_params->verified_status) . "\r\n\r\n" . $order_text;
			$email->body = $body;

			// Order update :
			$this->modifyOrder($dbOrder->order_id, $this->payment_params->verified_status, $history, $email);

			// Redirect customer to thank you page :
			$EndUrl = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=checkout&task=after_end&order_id='.$dbOrder->order_id.$this->url_itemid;
			$app->redirect($EndUrl);
		}
	}

	/**
	 * Set default backend option
	 */
	public function getPaymentDefaultValues(&$element) {
		$element->payment_name = 'Pay With Amazon';
		$element->payment_description = 'You can pay by credit card using this payment method';
		$element->payment_images = 'MasterCard,VISA,Credit_card,American_Express';

		$element->payment_params->sellerNote = 'Thank you for your order';
		$element->payment_params->color = 'Gold';
		$element->payment_params->notification = 1;
		$element->payment_params->testmode = 0;
		$element->payment_params->debug = 0;

		$element->payment_params->invalid_status = 'cancelled';
		$element->payment_params->pending_status = 'created';
		$element->payment_params->verified_status = 'confirmed';
	}

	/**
	 * Request processing
	 */
	protected function getArrayParams($action, $dbOrder, $context) {
		$params = array(
			// API REQUEST CALL :
			'Action' => $action,

			// params used for ALL :
			'AWSAccessKeyId' => trim($this->payment_params->accessKey),
			'SellerId' => trim($this->payment_params->sellerId),
			'AmazonOrderReferenceId' => $context['reference_id'],
			'SignatureMethod' => 'HmacSHA256',
			'SignatureVersion' => '2',
			'Timestamp' => gmdate("Y-m-d\TH:i:s.\\0\\0\\0\\Z", time()),
			'Version' => '2013-01-01',
			'MWSAuthToken' => $context['access_token']
		);

		// For SetOrderReferenceDetails
		if ($action == 'SetOrderReferenceDetails') {
			$params['OrderReferenceAttributes.OrderTotal.Amount'] = round($dbOrder->order_full_price, 2);
			$params['OrderReferenceAttributes.OrderTotal.CurrencyCode'] = $context['currency']->currency_code;
		}

		// For ConfirmOrderReference :
		// Nothing to add !

		// For Authorize :
		if ($action == 'Authorize') {
			$params['AuthorizationAmount.Amount'] = round($dbOrder->order_full_price, 2);
			$params['AuthorizationAmount.CurrencyCode'] = $context['currency']->currency_code;
			$params['CaptureNow'] = '1';
			// $params['TransactionTimeout'] = '0';
			$params['SellerAuthorizationNote'] = trim($this->payment_params->sellerNote);
			$params['AuthorizationReferenceId'] = $dbOrder->order_id; // TODO check this value.
			$params['ReturnURL'] = HIKASHOP_LIVE.'index.php?option=com_hikashop&ctrl=checkout&task=notify&notif_payment='.$this->name;
		}

		// For Closed :
		if ($action == 'CloseOrderReference') {
			$params['ClosureReason'] = 'Order finished';
		}
		//Debug check:
		if($this->payment_params->debug) {
			$this->writeToLog($action ."\n\n");
			$this->writeToLog("Data send to Amazon (just before request) : \n\n");
			$this->writeToLog(print_r($params,true));
		}

		return $params;
	}

	protected function getSignedData($params, $secretKey = null) {
		if(empty($secretKey))
			$secretKey = trim($this->payment_params->secretAccessKey);

		// Parameters need to be sorted (for signature)
		uksort($params, 'strcmp');

		// process & add signature :
		$params['Signature'] = $this->createSign($params, $secretKey);

		// Create a string (with urlencode :
		$stringData = $this->build_query($params, true);

		return $stringData;
	}

	protected function setUrlArray($amazonUrl) {
		// Set Amazon Url before request (Curl or socket) :
		$url = parse_url($amazonUrl);

		if(!isset($url['query']))
			$url['query'] = '';

		if(!isset($url['port'])) {
			if(!empty($url['scheme']) && in_array($url['scheme'], array('https', 'ssl'))) {
				$url['port'] = 443;
			} else {
				$url['port'] = 80;
			}
		}

		if(!empty($url['scheme']) && in_array($url['scheme'], array('https', 'ssl'))) {
			$url['host_socket'] = 'ssl://' . $url['host'];
		} else {
			$url['host_socket'] = $url['host'];
		}

		return $url;
	}

	/**
	 *
	 */
	protected function sendRequest($url, $data){
		$response = $this->_sendRequestSocket($url,$data);
		if(!$response) {
			$response = $this->_sendRequestCURL($url,$data);
		}
		return $response;
	}

	/**
	 *
	 */
	protected function _sendRequestCURL($url, $data) {
		// TEST CURL REQUEST PROCESS :
		if(!function_exists('curl_version')) {
			if($this->payment_params->debug)
				echo 'CURL is not available'. "\r\n\r\n";
			 return false;
		}

		$uri = $url['scheme'] . '://' . $url['host'] . $url['path'] . ($url['query'] != '' ? '?' . $url['query'] : '');
		// Curl to Amazon Servers :
		$ch = curl_init();

		if(!$ch){
			if($this->payment_params->debug)
				echo 'CURL could not be initialized'. "\r\n\r\n";
			 return false;
		}

		curl_setopt($ch, CURLOPT_URL, $uri);
		curl_setopt($ch, CURLOPT_PORT, $url['port']);
		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, true);
		curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 2);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

		curl_setopt($ch, CURLOPT_POST, true);
		curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
		curl_setopt($ch, CURLOPT_HEADER, false);

		$response = curl_exec($ch);

		//Debug check:
		if($this->payment_params->debug) {
			$this->writeToLog("Data recieved from Amazon (after curl) :\n\n ");
			$this->writeToLog(print_r($response,true));
		}

		$response = curl_exec($ch);

		$errno = curl_errno($ch);
		$error = curl_error($ch);

		curl_close($ch);

		if($errno){
			if($this->payment_params->debug) {
				echo 'CURL error number: '.$errno. "\r\n\r\n";
				echo 'CURL error message: '.$error. "\r\n\r\n";
			}
		}
		if (!$response) {
			if($this->payment_params->debug)
				echo 'CURL request didn\t return any data'. "\r\n\r\n";
			return false;
		}
		if($this->payment_params->debug) {
			$this->writeToLog('From Curl request process : /n/n');
			echo print_r($response, true) . "\r\n\r\n";
		}
		return $response;
	}

	/**
	 *
	 */
	protected function _sendRequestSocket($url, $data) {
		if(!function_exists('fsockopen')) {
			if($this->payment_params->debug)
				$this->writeToLog ("fsockopen function does not exist \r\n\r\n");
			return false;
		}
		if(!is_callable('fsockopen')) {
			if($this->payment_params->debug)
				$this->writeToLog ("fsockopen function is not callable \r\n\r\n");
			return false;
		}

		$fp = fsockopen($url['host_socket'], $url['port'], $errno, $errstr, 30);
		if(!$fp) {
			if($this->payment_params->debug)
				$this->writeToLog ("fsockopen connection couldn\'t be established \r\n\r\n");
			return false;
		}

		$uri = $url['path'] . ($url['query'] != '' ? '?' . $url['query'] : '');
		$header = 'POST '.$uri.' HTTP/1.1'."\r\n".
			'Host: '.$url['host']."\r\n".
			'Content-Type: application/x-www-form-urlencoded'."\r\n".
			'Content-Length: '.strlen($data)."\r\n".
			'Connection: close'."\r\n\r\n";

		fwrite($fp, $header . $data);
		$response = '';
		while(!feof($fp)) {
			$response .= fgets($fp, 1024);
		}
		fclose ($fp);

		if(empty($response)){
			if($this->payment_params->debug)
				$this->writeToLog ("fsockopen request didn\t return any data \r\n\r\n");
			return false;
		}

		// Remove response header
		$response = substr($response, strpos($response, "\r\n\r\n") + strlen("\r\n\r\n"));

		// Error response ?
		$errorCheck = strpos ($response, '<Error>');

		if ($errorCheck != false) {
			$errorObject = simplexml_load_string($response);
			$error = (string)$errorObject->Message;

			if ($this->payment_params->debug) {
				$this->writeToLog ("From Amazon server : ");
				$this->writeToLog($error);
				$this->writeToLog('Error with this data : ');
				$this->writeToLog($data);
				$this->writeToLog('Recieved data from AMazon : ');
				$this->writeToLog($response);
			}
			return false;
		}
		return $response;
	}

	/**
	 *
	 */
	protected function build_query($params, $encode = false) {
		$ret = array();
		foreach($params as $k => $v) {
			if($encode)
				$v = str_replace('%7E', '~', rawurlencode($v));
			$ret[] = $k . '=' . $v;
		}
		return implode('&', $ret);
	}

	/**
	 *
	 */
	protected function createSign($parameters, $secret) {
		//Amazon url from owner account zone;
		switch($this->payment_params->account_zone) {
			case 'ue':
			case 'uk':
			case 'de':

				$url = 'mws-eu.amazonservices.com';
				break;

			case 'us':

				$url = 'mws.amazonservices.com';
				break;

			case 'jp':

				$url = 'mws.amazonservices.jp';
				break;

			default:

				$url = 'mws.amazonservices.com';
				break;
		}
		$environnement = ($this->payment_params->environnement == 'sandbox') ? 'OffAmazonPayments_Sandbox' : 'OffAmazonPayments';
		$path = '/'. $environnement . '/' . '2013-01-01';

		//
		$data = 'POST'."\n".
			$url."\n".
			$path."\n";

		$parameters_str = is_string($parameters) ? $parameters : $this->build_query($parameters, true);

		// Sign string (data + encoded parameter)
		return base64_encode(hash_hmac('sha256', $data . $parameters_str, $secret, true));
	}

	/**
	 *
	 */
	protected function AmazonBtn ($vars) {
		// LOG WITH AMAZON BUTTON :
		return '
<div id="LoginWithAmazon"></div>
<script>
window.onAmazonLoginReady = function(){ amazon.Login.setClientId("'.$vars['lwaClientId'].'"); };
window.onAmazonPaymentsReady = function(){
	var authRequest;
	OffAmazonPayments.Button("LoginWithAmazon", "'.$vars['sellerId'].'", {
		type: "LwA", color:"'.$vars['color'].'", size:"'.$vars['size'].'", language:"'.$vars['language'].'",
		authorization: function(){
			loginOptions = {scope: "profile", popup: '.$vars['redirectMode'].'};
			authRequest = amazon.Login.authorize (loginOptions, "'.$vars['returnUrl'].'");
			},
		onError: function (error) {
            alert("Something went wrong, please try again !");
		}
	});
	document.getElementById(\'OffAmazonPaymentsWidgets0\').click();
};
</script>
<script async="async" src="'.$vars['amazonUrl'].'"></script>';

	}

	/*
	 *
	 */
	protected function _wallet($vars) {

		$wallet = ''.
		'<h3>Please choose your payment method : </h3>'.
		'<div id="walletWidgetDiv"></div>'.
		'<script>'.
			"window.onAmazonLoginReady = function() {amazon.Login.setClientId('".$vars['lwaClientId']."'); };".
			'window.onAmazonPaymentsReady = function() {'.
				'var orderReferenceId = null;'.

				'new OffAmazonPayments.Widgets.Wallet({ '.
					"sellerId: '".$vars['sellerId']."', ".
					'onOrderReferenceCreate: function(orderReference) { '.
						'orderReferenceId = orderReference.getAmazonOrderReferenceId(); '.
					'},'.
					'design: {'.
						"designMode: 'responsive'".
					'},'.
					'onPaymentSelect: function(orderReference) { '.
						"document.getElementById('reference_id').value = orderReferenceId;".
					'},'.
					'onError: function(error) {'.
						'alert(" Something went wrong !");'.
						"console.log(error.getErrorCode() + ': ' + error.getErrorMessage());".
					'}'.
				'}).bind("walletWidgetDiv");'.
			'}'.
		'</script>'.
		'<script async="=async" type="text/javascript" src="'.$vars['amazonUrl'].'"></script>'.
		'<form method="post" action="'.$vars['returnNotif'].'">
			<input type="hidden" name="access_token" value="'.$vars['access_token'].'" />
			<input type="hidden" name="reference_id" id="reference_id" value="" />
			<input type="hidden" name="payment_selected" id="paymentflag" value="true" />
			<div class="hikashop_checkout_buttons">
				<div class="buttons_right">
					<input class="hikabtn hikabtn_checkout_next" type="submit" value="Confirm" />
				</div>
				<div style="clear:both"></div>
			</div>
		</form>';

		return $wallet;
	}
}
